@@ -13,30 +13,51 @@
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 
 (texmacs-module (convert latex tmtex)
-  (:use
-    (convert tools tmpre) (convert tools tmtable)
-    (convert rewrite tmtm-eqns) (convert rewrite tmtm-brackets)
-    (convert latex texout))
-  (:export
-    texmacs->latex
-    tmtex tmtex-initialize)) ;; for tmtex-test
+  (:use (convert tools tmpre)
+	(convert tools old-tmtable)
+	(convert tools tmlength)
+	(convert rewrite tmtm-eqns)
+	(convert rewrite tmtm-brackets)
+	(convert latex texout)
+	(convert latex latex-tools)))
 
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;; Initialization
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 
 (define tmtex-env (make-ahash-table))
+(define tmtex-serial 0)
+(define tmtex-image-root-url (string->url "image"))
+(define tmtex-image-root-string "image")
 (define tmtex-appendices? #f)
-(define tmtex-faithful-style? #f)
+(define tmtex-replace-style? #t)
 (define tmtex-indirect-bib? #f)
+(define tmtex-oriental? #f)
+(define tmtex-chinese? #f)
+(define tmtex-japanese? #f)
+(define tmtex-korean? #f)
+(define tmtex-taiwanese? #f)
 
-(define (tmtex-initialize opts)
+(tm-define (tmtex-initialize opts)
+  (set! tmtex-env (make-ahash-table))
+  (set! tmtex-serial 0)
+  (if (== (url-suffix current-save-target) "tex")
+      (begin
+	(set! tmtex-image-root-url (url-unglue current-save-target 4))
+	(set! tmtex-image-root-string
+	      (url->string (url-tail tmtex-image-root-url))))
+      (begin
+	(set! tmtex-image-root-url (string->url "image"))
+	(set! tmtex-image-root-string "image")))
   (set! tmtex-appendices? #f)
-  (set! tmtex-faithful-style?
-	(== (assoc-ref opts "texmacs->latex:faithful-style") "on"))
+  (set! tmtex-replace-style?
+	(== (assoc-ref opts "texmacs->latex:replace-style") "on"))
   (set! tmtex-indirect-bib?
 	(== (assoc-ref opts "texmacs->latex:indirect-bib") "on"))
-  (set! tmtex-env (make-ahash-table)))
+  (set! tmtex-use-catcodes?
+	(== (assoc-ref opts "texmacs->latex:use-catcodes") "on"))
+  (set! tmtex-use-macros?
+	(== (assoc-ref opts "texmacs->latex:use-macros") "on")))
 
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;; Data
@@ -51,21 +72,16 @@
   (det ((left|) "c" (right|) #f))
   (choice ((left\{) "l" (right.) #f)))
 
-(drd-group tex-mathops%
-  arccos arcsin arctan cos cosh cot coth csc deg det dim exp gcd
-  hom inf ker lg lim liminf limsup ln log max min Pr sec sin
-  sinh sup tan tanh)
-
 (drd-table tex-with-cmd%
-  (("font-family" "rm") textrm)
-  (("font-family" "ss") textsf)
-  (("font-family" "tt") texttt)
-  (("font-series" "medium") textmd)
-  (("font-series" "bold") textbf)
-  (("font-shape" "right") textup)
-  (("font-shape" "slanted") textsl)
-  (("font-shape" "italic") textit)
-  (("font-shape" "small-caps") textsc)
+  (("font-family" "rm") tmtextrm)
+  (("font-family" "ss") tmtextsf)
+  (("font-family" "tt") tmtexttt)
+  (("font-series" "medium") tmtextmd)
+  (("font-series" "bold") tmtextbf)
+  (("font-shape" "right") tmtextup)
+  (("font-shape" "slanted") tmtextsl)
+  (("font-shape" "italic") tmtextit)
+  (("font-shape" "small-caps") tmtextsc)
   (("math-font" "cal") mathcal)
   (("math-font" "cal*") mathscr)
   (("math-font" "cal**") EuScript)
@@ -128,7 +144,7 @@
 
 (define (tmtex-env-reset var)
   (let ((val (tmtex-env-list var)))
-    (if (not (null? val))
+    (if (nnull? val)
 	(ahash-set! tmtex-env var (cdr val)))))
 
 (define (tmtex-env-assign var val)
@@ -142,10 +158,13 @@
 (define (tex-concat-similar l)
   (if (or (null? l) (null? (cdr l))) l
       (let ((r (tex-concat-similar (cdr l))))
-	    (if (and (func? (car l) '!sup) (func? (car r) '!sup))
-		(cons (list '!sup (tex-concat (list (cadar l) (cadar r))))
-		      (cdr r))
-		(cons (car l) r)))))
+	(cond ((and (func? (car l) '!sub) (func? (car r) '!sub))
+	       (cons (list '!sub (tex-concat (list (cadar l) (cadar r))))
+		     (cdr r)))
+	      ((and (func? (car l) '!sup) (func? (car r) '!sup))
+	       (cons (list '!sup (tex-concat (list (cadar l) (cadar r))))
+		     (cdr r)))
+	      (else (cons (car l) r))))))
 
 (define (tex-concat-list l)
   (cond ((null? l) l)
@@ -199,8 +218,30 @@
 	((string-starts? s "b-cal-")
 	 (tex-apply 'tmmathbf (tmtex-modified-token 'mathcal s 6)))
 	((string-starts? s "b-") (tmtex-modified-token 'tmmathbf s 2))
+	((and (string-starts? s "#") tmtex-oriental?)
+	 (cond (tmtex-japanese?
+		(let* ((qs (string-append "<" s ">"))
+		       (cv (string-convert qs "Cork" "ISO-2022-JP"))
+		       (ex (list->string (list #\33 #\50 #\102))))
+		  (set! cv (string-append cv ex))
+		  (list '!widechar (string->symbol cv))))
+	       (tmtex-korean?
+		(let* ((qs (string-append "<" s ">"))
+		       (cv (string-convert qs "Cork" "UTF-8")))
+		  (list '!widechar (string->symbol cv))))
+	       ((or tmtex-chinese? tmtex-taiwanese?)
+		(let* ((qs (string-append "<" s ">"))
+		       ;;(cv (string-convert qs "Cork" "cp936")) ; Chinese?
+		       ;;(cv (string-convert qs "Cork" "cp950")) ; Taiwanese ?
+		       (cv (string-convert qs "Cork" "UTF-8")))
+		  (set! cv (string-append cv ex))
+		  (list '!widechar (string->symbol cv))))))
 	(else (let ((ss (list (string->symbol s))))
-		(if group? (list '!group ss) ss)))))
+		(cond ((not (drd-in? (car ss) latex-symbol%))
+		       (display* "TeXmacs] non converted symbol: " s "\n")
+		       "")
+		      (group? (list '!group ss))
+		      (else (list '!symbol ss)))))))
 
 (define (tmtex-token l routine group?)
   (receive (p1 p2) (list-break (cdr l) (lambda (x) (== x #\>)))
@@ -218,10 +259,17 @@
 (define (tmtex-break-char? c)
   (string-index "+ -:=,?;()[]{}<>/" c))
 
+(define (tmtex-text-list-space l)
+  (cond ((null? l) l)
+	((== (car l) #\space)
+	 (cons (list (string->symbol " ")) (tmtex-text-list-space (cdr l))))
+	(else (tmtex-text-list l))))
+
 (define (tmtex-text-list l)
   (if (null? l) l
       (let ((c (car l)))
 	(cond ((== c #\<) (tmtex-token l tmtex-text-list #t))
+	      ((== c #\space) (cons c (tmtex-text-list-space (cdr l))))
 	      ((tmtex-special-char? c)
 	       (cons (list (string->symbol (char->string c)))
 		     (tmtex-text-list (cdr l))))
@@ -240,8 +288,8 @@
   (receive (p q) (list-break l (lambda (c) (not (char-alphabetic? c))))
     (let* ((op (list->string p))
 	   (tail (tmtex-math-list q)))
-      (if (drd-in? (string->symbol op) tex-mathops%)
-	  (cons (tex-apply (string->symbol op)) tail)
+      (if (drd-in? (string->symbol op) latex-operator%)
+	  (cons (list '!symbol (tex-apply (string->symbol op))) tail)
 	  (cons (tex-apply 'tmop op) tail)))))
 
 (define (tmtex-math-list l)
@@ -259,24 +307,11 @@
 	      ((== c #\*) (tmtex-math-list (cdr l)))
 ;;	      ((== c #\space) (tmtex-math-list (cdr l)))
 	      ((and (char-alphabetic? c)
-		    (not (null? (cdr l)))
+		    (nnull? (cdr l))
 		    (char-alphabetic? (cadr l)))
 	       (tmtex-math-operator l))
 	      (else (cons c (tmtex-math-list (cdr l))))))))
 
-(define (tmtex-no-math-space c1 c2)
-  (or (== c2 #\,)
-      (and (== c1 c2) (in? c1 '(#\( #\) #\[ #\])))
-      (and (char? c1) (char? c2)
-	   (char-numeric? c1) (char-numeric? c2))))
-
-(define (tmtex-math-spaces l)
-  (if (or (null? l) (null? (cdr l))) l
-      (let ((tail (tmtex-math-spaces (cdr l))))
-	(if (tmtex-no-math-space (car l) (cadr l))
-	    (cons (car l) tail)
-	    (cons* (car l) #\space tail)))))
-
 (define (tmtex-verb-list l)
   (if (null? l) l
       (let ((c (car l)))
@@ -306,8 +341,8 @@
 (define (tmtex-string s)
   (let* ((l (string->list s))
 	 (t (if (tmtex-math-mode?)
-            (tmtex-math-list l)
-            (tmtex-text-list l)))
+		(tmtex-math-list l)
+		(tmtex-text-list l)))
 	 (r (tmtex-string-produce t)))
     (tex-concat r)))
 
@@ -327,8 +362,11 @@
 	((== x "seminar") "slides")
 	((in? x '("tmarticle" "tmdoc" "mmxdoc")) "article")
 	((in? x '("tmbook" "tmmanual")) "book")
-	((in? x '("acmconf" "amsart" "jsc" "svjour")) x)
-	(tmtex-faithful-style? x)
+	;;((in? x '("acmconf" "amsart" "svjour" "elsart")) x)
+	;;((in? x '("jsc")) "elsart")
+	((in? x '("acmconf" "amsart")) x)
+	((in? x '("svjour" "elsart" "jsc")) "article")
+	((not tmtex-replace-style?) x)
 	(else #f)))
 
 (define (tmtex-filter-styles l)
@@ -337,39 +375,49 @@
 	     (tail (tmtex-filter-styles (cdr l))))
 	(if next (cons next tail) tail))))
 
+(define (macro-definition? x)
+  (and (func? x 'assign 2)
+       (string? (cadr x))
+       (func? (caddr x) 'macro)))
+
 (define (tmtex-filter-preamble l)
-  (define (append-lists l)
-    (if (null? l) l (append (car l) (append-lists (cdr l)))))
-  (cond ((or (not (list? l)) (null? l)) '())
-	((== (car l) 'assign) (list l))
-	(else (append-lists (map tmtex-filter-preamble (cdr l))))))
+  (cond ((or (nlist? l) (null? l)) '())
+	((macro-definition? l) (list l))
+	((== (car l) 'hide-preamble) (cdadr l))
+	(else (append-map tmtex-filter-preamble (cdr l)))))
 
 (define (tmtex-filter-body l)
-  (cond ((or (not (list? l)) (null? l)) l)
+  (cond ((or (nlist? l) (null? l)) l)
 	((== (car l) 'assign) "")
+	((== (car l) 'hide-preamble) "")
 	(else (cons (car l) (map tmtex-filter-body (cdr l))))))
 
 (define (tmtex-file l)
   (let* ((doc (car l))
-	 (styles (cdadr l))
+	 (styles (cadr l))
 	 (lang (caddr l))
 	 (init (cadddr l))
 	 (doc-preamble (tmtex-filter-preamble doc))
 	 (doc-body (tmtex-filter-body doc)))
+    (if (== (get-preference "texmacs->latex:expand-user-macros") "on")
+	(set! doc-preamble '()))
     (if (null? styles) (tmtex doc)
-	(begin
-	  (list '!file
-		(tmtex doc-body)
-		(tmtex-filter-styles styles)
-		lang
-		init
-		(map-in-order tmtex doc-preamble))))))
+	(let* ((body* (tmtex doc-body))
+	       (styles* (tmtex-filter-styles styles))
+	       (preamble* (ahash-with tmtex-env :preamble #t
+			    (map-in-order tmtex doc-preamble))))
+	  (list '!file body* styles* lang init preamble*)))))
 
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;; Simple text
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 
 (define (tmtex-noop l) "")
+(define (tmtex-default s l) (cons (string->symbol s) (tmtex-list l)))
+(define (tmtex-id l) (tmtex (car l)))
+(define (tmtex-second l) (tmtex (cadr l)))
+(define (tmtex-hide-part s l) "")
+(define (tmtex-show-part s l) (tmtex (cadr l)))
 
 (define (tmtex-document l)
   (cons '!document (tmtex-list l)))
@@ -398,7 +446,7 @@
 (define (tmtex-script? x)
   (or (func? x '!sub)
       (func? x '!sup)
-      (and (string? x) (not (== x "")) (in? (string-ref x 0) '(#\' #\,)))
+      (and (string? x) (!= x "") (in? (string-ref x 0) '(#\' #\, #\) #\])))
       (and (func? x '!concat) (tmtex-script? (cadr x)))))
 
 (define (tmtex-math-concat-spaces l)
@@ -409,10 +457,22 @@
 	    (cons head tail)
 	    (cons* head " " tail)))))
 
+(define (tmtex-rewrite-no-break l)
+  (cond ((null? l) l)
+	((and (string? (car l)) (string-ends? (car l) " ")
+	      (nnull? (cdr l)) (== (cadr l) '(no-break)))
+	 (let* ((s (substring (car l) 0 (- (string-length (car l)) 1)))
+		(r (tmtex-rewrite-no-break (cddr l))))
+	   (if (== s "") (cons '(!nbsp) r) (cons* s '(!nbsp) r))))
+	(else (cons (car l) (tmtex-rewrite-no-break (cdr l))))))
+
 (define (tmtex-concat l)
   (if (tmtex-math-mode?)
       (tex-concat (tmtex-math-concat-spaces (tmtex-list l)))
-      (tex-concat (tmtex-list l))))
+      (tex-concat (tmtex-list (tmtex-rewrite-no-break l)))))
+
+(define (tmtex-group l)
+  (tmtex-function '!group l))
 
 (define (tmtex-no-first-indentation l) (tex-apply 'noindent))
 (define (tmtex-line-break l) (tex-apply 'linebreak))
@@ -420,6 +480,7 @@
 (define (tmtex-new-page l) (tex-apply 'newpage))
 (define (tmtex-new-line l) (tex-apply '!newline))
 (define (tmtex-next-line l) (list '!nextline))
+(define (tmtex-no-break l) '(!group (nobreak)))
 
 (define (tmtex-decode-length s)
   ;; FIXME: should be completed
@@ -449,8 +510,15 @@
 (define (tmtex-space l)
   (tmtex-hspace (list (car l))))
 
-(define (tmtex-float-make size type position x caption)
-  (list 'tmfloat position size type (tmtex x) (tmtex caption)))
+(define (tmtex-float-make size type position x capt)
+  (let* ((body (tmtex x))
+	 (caption (tmtex capt))
+	 (body* `(!paragraph ,body (caption ,caption))))
+    (cond ((and (== size "big") (== type "figure"))
+	   `((!begin "figure" (!option ,position)) ,body*))
+	  ((and (== size "big") (== type "table"))
+	   `((!begin "table" (!option ,position)) ,body*))
+	  (else (list 'tmfloat position size type body caption)))))
 
 (define (tmtex-float-table? x)
   (or (func? x 'small-table 2) (func? x 'big-table 2)))
@@ -483,11 +551,8 @@
 ;; Mathematics
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 
-(define (tmtex-group l)
-  (tmtex-function '!group l))
-
 (define (tmtex-large-decode s)
-  (cond ((not (string? s)) ".")
+  (cond ((nstring? s) ".")
         ((in? s '("(" ")" "[" "]" "|" "/" ".")) s)
 	((== s "||") "\\|")
 	((== s "\\") "\\backslash")
@@ -506,7 +571,7 @@
   (list (string->symbol (string-append "right" (tmtex-large-decode (car l))))))
 
 (define (tmtex-big-decode s)
-  (cond ((not (string? s)) "bignone")
+  (cond ((nstring? s) "bignone")
         ((in? s '("sum" "prod" "int" "oint" "coprod")) s)
 	((== s "amalg") "coprod")
 	((== s "pluscup") "uplus")
@@ -548,11 +613,22 @@
 (define (tmtex-lsup l)
   (tmtex (list 'concat (list 'text "") (list 'rsup (car l)))))
 
+(define (tmtex-contains-table? x)
+  (cond ((nlist? x) #f)
+	((and (>= (length x) 2) (== (car x) '!table)) #t)
+	(else (list-or (map-in-order tmtex-contains-table? (cdr x))))))
+
+(define (tmtex-script which script)
+  (with r (tmtex script)
+    (if (tmtex-contains-table? r)
+	(list which (list 'tmscript r))
+	(list which r))))
+
 (define (tmtex-rsub l)
-  (tmtex-function '!sub l))
+  (tmtex-script '!sub (car l)))
 
 (define (tmtex-rsup l)
-  (tmtex-function '!sup l))
+  (tmtex-script '!sup (car l)))
 
 (define (tmtex-frac l)
   (tmtex-function 'frac l))
@@ -566,51 +642,76 @@
 
 (define (tmtex-token? s)
   (or (= (string-length s) 1)
-      (and (not (== s ""))
+      (and (!= s "")
 	   (== (string-ref s 0) #\<)
 	   (== (string-index s #\>) (- (string-length s) 1)))))
        
 (define (tmtex-wide-star? x)
   (cond ((func? x 'wide* 1) (tmtex-wide-star? (cadr x)))
-	((not (string? x)) #t)
+	((nstring? x) #t)
 	(else (not (tmtex-token? x)))))
 
 (define (tmtex-wide-star l)
-  (let ((wide (tmtex-wide-star? (car l))) (arg (tmtex (car l))) (acc (cadr l)))
-    (cond ((not (string? acc)) arg)
-	  ((== acc "<bar>") (list 'underline arg))
-	  ((== acc "<wide-bar>") (list 'underline arg))
-	  ((== acc "<wide-underbrace>") (list 'underbrace arg))
-	  ((== acc "<wide-overbrace*>")
-	    (tmtex-below (list (car l) (list 'text (list 'downbracefill)))))
+  (let ((wide (tmtex-wide-star? (car l)))
+	(arg (tmtex (car l)))
+	(acc (cadr l)))
+    (if (and (string? acc) (string-starts? acc "<wide-"))
+	(set! acc (string-append "<" (substring acc 6 (string-length acc)))))
+    (cond ((nstring? acc) arg)
 	  ((== acc "~")
 	   (tmtex-below (list (car l) (list 'mbox (list 'textasciitilde)))))
-	  (else (list 'wide arg)))))
+	  ((== acc "<bar>") (list 'underline arg))
+	  ((in? acc '("<underbrace>" "<underbrace*>"))
+	   (list 'underbrace arg))
+	  ((in? acc '("<overbrace>" "<overbrace*>"))
+	   (tmtex-below `(,(car l) (text (downbracefill)))))
+	  ;; imperfect translations
+	  ((in? acc '("<squnderbrace>" "<squnderbrace*>"))
+	   (list 'underbrace arg))
+	  ((in? acc '("<sqoverbrace>" "<sqoverbrace*>"))
+	   (tmtex-below `(,(car l) (text (downbracefill)))))
+	  (else
+	   (display* "TeXmacs] non converted accent below: " acc "\n")
+	   arg))))
 
 (define (tmtex-wide? x)
   (cond ((func? x 'wide 1) (tmtex-wide? (cadr x)))
-	((not (string? x)) #t)
+	((nstring? x) #t)
 	(else (not (tmtex-token? x)))))
 
 (define (tmtex-wide l)
-  (let ((wide (tmtex-wide? (car l))) (arg (tmtex (car l))) (acc (cadr l)))
-    (cond ((not (string? acc)) arg)
-	  ((== acc "<check>") (list 'check arg))
+  (let ((wide (tmtex-wide? (car l)))
+	(arg (tmtex (car l)))
+	(acc (cadr l)))
+    (if (and (string? acc) (string-starts? acc "<wide-"))
+	(set! acc (string-append "<" (substring acc 6 (string-length acc)))))
+    (cond ((nstring? acc) arg)
+	  ((in? acc '("<hat>" "^")) (list (if wide 'widehat 'hat) arg))
+	  ((in? acc '("<tilde>" "~")) (list (if wide 'widetilde 'tilde) arg))
+	  ((== (cadr l) "<wide-bar>") (list 'overline arg))
+	  ((== acc "<bar>") (list (if wide 'overline 'bar) arg))
 	  ((== acc "<vect>") (list (if wide 'overrightarrow 'vec) arg))
+	  ((== acc "<breve>") (list 'breve arg))
+	  ((== acc "<check>") (list 'check arg))
 	  ((== acc "<acute>") (list 'acute arg))
 	  ((== acc "<grave>") (list 'grave arg))
-	  ((== acc "<breve>") (list 'breve arg))
-	  ((== acc "<bar>") (list (if wide 'overline 'bar) arg))
 	  ((== acc "<dot>") (list 'dot arg))
 	  ((== acc "<ddot>") (list 'ddot arg))
 	  ((== acc "<dddot>") (list 'dddot arg))
 	  ((== acc "<ddddot>") (list 'ddddot arg))
-	  ((== acc "<wide-overbrace>") (list 'overbrace arg))
-	  ((== acc "<wide-underbrace*>")
-	    (tmtex-above (list (car l) (list 'text (list 'upbracefill)))))
-	  ((== acc "~") (list (if wide 'widetilde 'tilde) arg))
-	  ((== acc "^") (list (if wide 'widehat 'hat) arg))
-	  (else arg))))
+	  ((in? acc '("<overbrace>" "<overbrace*>"))
+	   (list 'overbrace arg))
+	  ((in? acc '("<underbrace>" "<underbrace*>"))
+	   (tmtex-above `(,(car l) (text (upbracefill)))))
+	  ;; FIXME: imperfect translations
+	  ((== acc "<abovering>") (list 'dot arg))
+	  ((in? acc '("<sqoverbrace>" "<sqoverbrace*>"))
+	   (list 'overbrace arg))
+	  ((in? acc '("<squnderbrace>" "<squnderbrace*>"))
+	   (tmtex-above `(,(car l) (text (upbracefill)))))
+	  (else
+	   (display* "TeXmacs] non converted accent: " acc "\n")
+	   arg))))
 
 (define (tmtex-neg l)
   (tmtex-function 'not l))
@@ -620,6 +721,9 @@
 	 (children (map (lambda (x) (list 'chunk (tmtex x))) (cdr l))))
     (list root (tex-concat children))))
 
+(define (tmtex-tree-eps l)
+  (tmtex-eps (cons 'tree l)))
+
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;; Tables
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
@@ -697,38 +801,59 @@
 	      (else #f)))
       (drd-ref tex-assign-cmd% (list var val))))
 
+(define (tmlength->texlength len)
+  ;; TODO: rewrite (quote x) -> x and (tmlen ...) -> ...pt
+  (with tmlen (string->tmlength (force-string len))
+    (if (tmlength-null? tmlen) "0pt"
+	(let* ((val (tmlength-value tmlen))
+	       (unit (symbol->string (tmlength-unit tmlen)))
+	       (val-string (number->string val)))
+	  (cond ((== unit "fn") (string-append val-string "em"))
+		(else len))))))
+
+(define (tmtex-make-parmod x y z arg)
+  (set! x (tmlength->texlength x))
+  (set! y (tmlength->texlength y))
+  (set! z (tmlength->texlength z))
+  (if (and (tmlength-zero? (string->tmlength x))
+	   (tmlength-zero? (string->tmlength y))
+	   (tmlength-zero? (string->tmlength z)))
+      arg
+      (list (list '!begin "tmparmod" x y z) arg)))
+
 (define (tmtex-with-one var val arg)
   (if (== var "mode")
       (let ((old (tmtex-env-get-previous "mode")))
-	(cond ((and (== val "text") (not (== old "text")))
+	(cond ((and (== val "text") (!= old "text"))
 	       (list 'text arg))
-	      ((and (== val "math") (not (== old "math")))
+	      ((and (== val "math") (!= old "math")
+		    (ahash-ref tmtex-env :preamble))
+	       (list 'ensuremath arg))
+	      ((and (== val "math") (!= old "math"))
 	       (list '!math arg))
 	      (else arg)))
       (let ((w (tmtex-get-with-cmd var val))
 	    (a (tmtex-get-assign-cmd var val)))
 	(cond (w (list w arg))
 	      (a (list '!group (tex-concat (list (list a) " " arg))))
-	      ((== "par-left" var)
-               (list (list '!begin "tmparmod" val "0pt" "0pt") arg))
-	      ((== "par-right" var)
-               (list (list '!begin "tmparmod" "0pt" val "0pt") arg))
-	      ((== "par-first" var)
-               (list (list '!begin "tmparmod" "0pt" "0pt" val) arg))
+	      ((== "par-left" var) (tmtex-make-parmod val "0pt" "0pt" arg))
+	      ((== "par-right" var) (tmtex-make-parmod "0pt" val "0pt" arg))
+	      ((== "par-first" var) (tmtex-make-parmod "0pt" "0pt" val arg))
 	      ((== var "color")
 	      	(list '!group (tex-concat (list (list 'color val) " " arg))))
 	      (else arg)))))
 
 (define (tmtex-with l)
-  (if (null? l) ""
-      (if (null? (cdr l)) (tmtex (car l))
-	  (let ((var (force-string (car l)))
-		(val (force-string (cadr l)))
-		(next (cddr l)))
-	    (tmtex-env-set var val)
-	    (let ((r (tmtex-with-one var val (tmtex-with next))))
-	      (tmtex-env-reset var)
-	      r)))))
+  (cond ((null? l) "")
+	((null? (cdr l)) (tmtex (car l)))
+	((func? (cAr l) 'graphics) (tmtex-eps (cons 'with l)))
+	(else (let ((var (force-string (car l)))
+		    (val (force-string (cadr l)))
+		    (next (cddr l)))
+		(tmtex-env-set var val)
+		(let ((r (tmtex-with-one var val (tmtex-with next))))
+		  (tmtex-env-reset var)
+		  r)))))
 
 (define (tmtex-var-name-sub l)
   (if (null? l) l
@@ -741,7 +866,7 @@
 	      (else r)))))
 
 (define (tmtex-var-name var)
-  (cond ((not (string? var)) "")
+  (cond ((nstring? var) "")
 	((drd-in? (string->symbol var) tmtex-protected%)
 	 (string-append "tm" var))
 	((<= (string-length var) 1) var)
@@ -763,15 +888,17 @@
 	    (tmtex-args-sub (cdr l) args))))
 
 (define (tmtex-args x args)
-  (cond ((not (list? x)) x)
+  (cond ((nlist? x) x)
 	((or (func? x 'arg) (func? x 'value))
 	 (let ((n (tmtex-args-search (cadr x) args)))
 	   (if n (list '!arg (number->string n)) (tmtex-args-sub x args))))
 	(else (tmtex-args-sub x args))))
 
 (define (tmtex-assign l)
-  (let ((var (tmtex-var-name (car l))) (val (cadr l)))
-    (if (not (== var ""))
+  (let ((var (tmtex-var-name (car l)))
+	(val (cadr l)))
+    (while (func? val 'quote 1) (set! val (cadr val)))
+    (if (!= var "")
 	(begin
 	  (tmtex-env-assign var val)
 	  (cond ((string? val)
@@ -794,6 +921,9 @@
 ;; Other primitives
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 
+(define (tmtex-quote l)
+  (tmtex (car l)))
+
 (define (tmtex-label l)
   (list 'label (force-string (car l))))
 
@@ -804,16 +934,46 @@
   (list 'pageref (force-string (car l))))
 
 (define (tmtex-specific l)
-  (if (== (car l) "latex") (tmtex (cadr l)) ""))
+  (cond ((== (car l) "latex") (tmtex (cadr l)))
+	((== (car l) "image") (tmtex-eps (cadr l)))
+	(else "")))
 
 (define (tmtex-hyperlink l)
   (tmtex-function 'tmhlink l))
 
 (define (tmtex-action l)
-  (tmtex-function 'tmaction l))
+  (list 'tmaction (tmtex (car l)) (tmtex (cadr l))))
+
+(define (tmtex-eps-names)
+  (set! tmtex-serial (+ tmtex-serial 1))
+  (let* ((postfix (string-append "-" (number->string tmtex-serial) ".eps"))
+	 (name-url (url-glue tmtex-image-root-url postfix))
+	 (name-string (string-append tmtex-image-root-string postfix)))
+    (values name-url name-string)))
+
+(define (tmtex-eps x)
+  (if (tmtex-math-mode?) (set! x `(with "mode" "math" ,x)))
+  (receive (name-url name-string) (tmtex-eps-names)
+    (print-snippet name-url x)
+    (list 'epsfig (string-append "file=" name-string))))
+
+(define (tmtex-graphics l)
+  (tmtex-eps (cons 'graphics l)))
+
+(define (tmtex-as-eps name)
+  (with u (url-relative current-save-target (string->url name))
+    (if (or (string-ends? name ".ps")
+	    (string-ends? name ".eps")
+	    (not (url-exists? u)))
+	(list 'epsfig (string-append "file=" name))
+	(let* ((suffix (url-suffix u))
+	       (fm (string-append (format-from-suffix suffix) "-file")))
+	  (receive (name-url name-string) (tmtex-eps-names)
+	    (convert-to-file u fm "postscript-file" name-url)
+	    (list 'epsfig (string-append "file=" name-string)))))))
 
 (define (tmtex-postscript l)
-  (let* ((fig (list 'epsfig (string-append "file=" (force-string (car l)))))
+  (let* ((fig (tmtex-as-eps (force-string (car l))))
 	 (hor (if (== (cadr l) "") "!" (tmtex-decode-length (cadr l))))
 	 (ver (if (== (caddr l) "") "!" (tmtex-decode-length (caddr l)))))
     (if (or (string-starts? hor "*") (string-starts? hor "/")) (set! hor "!"))
@@ -841,7 +1001,7 @@
 		      (else (append (car l) (list sep) r)))))))
 
 (define (tmtex-select-data expr tag)
-  (let* ((data (tm-select expr (list tag)))
+  (let* ((data (select expr (list tag)))
 	 (sep (if (== tag 'author-address) '(!nextline) "; "))
 	 (fun (lambda (x)
 		(cond ((func? x 'document)
@@ -871,7 +1031,7 @@
 (define (tmtex-doc-data s l)
   (let* ((tag (cons s l))
 	 (title (tmtex-select-data tag 'doc-title))
-	 (authors (map tmtex-make-author (tm-select tag '(doc-author-data))))
+	 (authors (map tmtex-make-author (select tag '(doc-author-data))))
 	 (date (tmtex-select-data tag 'doc-date))
 	 (note (tmtex-select-data tag 'doc-note))
 	 (keywords (tmtex-select-data tag 'doc-keywords))
@@ -890,14 +1050,15 @@
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 
 (define (tmtex-std-env s l)
+  (if (== s "quote-env") (set! s "quote"))
   (list (list '!begin s) (tmtex (car l))))
 
 (define (tmtex-appendix s l)
-  (if tmtex-appendices?
-      (list 'chapter (tmtex (car l)))
+  (with app (list (if (latex-book-style?) 'chapter 'section) (tmtex (car l)))
+    (if tmtex-appendices? app
       (begin
 	(set! tmtex-appendices? #t)
-	(list '!concat '(appendix) (list 'chapter (tmtex (car l)))))))
+	(list '!concat '(appendix) app)))))
 
 (define (tmtex-tt-document l)
   (cond ((null? l) "")
@@ -917,7 +1078,8 @@
 (define (tmtex-verbatim s l)
   (if (func? (car l) 'document)
       (list '!verbatim (tmtex-tt (car l)))
-      (list '!verb (tmtex-tt (car l)))))
+      (list 'tmtexttt (tmtex (car l)))))
+;;(list '!verb (tmtex-tt (car l)))))
 
 (define (tmtex-list-env s l)
   (let* ((r (string-replace s "-" ""))
@@ -926,6 +1088,30 @@
 		  (else r))))
     (list (list '!begin t) (tmtex (car l)))))
 
+(define (tmtex-tiny s l)
+  (tex-apply 'tiny (tmtex (car l))))
+
+(define (tmtex-scriptsize s l)
+  (tex-apply 'scriptsize (tmtex (car l))))
+
+(define (tmtex-footnotesize s l)
+  (tex-apply 'footnotesize (tmtex (car l))))
+
+(define (tmtex-normalsize s l)
+  (tex-apply 'normalsize (tmtex (car l))))
+
+(define (tmtex-large s l)
+  (tex-apply 'large (tmtex (car l))))
+
+(define (tmtex-Large s l)
+  (tex-apply 'Large (tmtex (car l))))
+
+(define (tmtex-LARGE s l)
+  (tex-apply 'LARGE (tmtex (car l))))
+
+(define (tmtex-Huge s l)
+  (list 'Huge (tmtex (car l))))
+
 (define (tmtex-equation s l)
   (tmtex-env-set "mode" "math")
   (let ((r (tmtex (car l))))
@@ -940,6 +1126,9 @@
     (tmtex-env-reset "mode")
     r))
 
+(define (tmtex-math s l)
+  (tmtex `(with "mode" "math" ,(car l))))
+
 (define (tmtex-dummy s l)
   "")
 
@@ -952,6 +1141,7 @@
 	 (let* ((l (cadr (caddr doc)))
 		(s (if (string-starts? l "bib-") (string-drop l 4) l)))
 	   (cons* 'concat (list 'bibitem* (cadadr doc) s) (cdddr doc))))
+	((func? doc 'bib-list 2) (tmtex-bib-sub (cAr doc)))
 	(else (map tmtex-bib-sub doc))))
 
 (define (tmtex-bib-max l)
@@ -1014,7 +1204,7 @@
 
 (define (tmtex-cite-list l)
   (cond ((null? l) "")
-	((not (string? (car l))) (tmtex-cite-list (cdr l)))
+	((nstring? (car l)) (tmtex-cite-list (cdr l)))
 	((null? (cdr l)) (car l))
 	(else (string-append (car l) "," (tmtex-cite-list (cdr l))))))
 
@@ -1039,6 +1229,9 @@
 (define (tmtex-menu s l)
   (tex-concat (cons (tmtex-menu-one (car l)) (tmtex-menu-list (cdr l)))))
 
+(define ((tmtex-rename into) s l)
+  (tmtex-apply into (tmtex-list l)))
+
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;; The main conversion routines
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
@@ -1046,6 +1239,7 @@
 (define (tmtex-apply key args)
   (let ((n (length args))
 	(r (drd-ref tmtex-methods% key)))
+    (if (in? key '(quote quasiquote unquote)) (set! r tmtex-noop))
     (if r (r args)
 	(let ((p (drd-ref tmtex-tmstyle% key)))
 	  (if (and p (or (= (cadr p) -1) (= (cadr p) n)))
@@ -1071,7 +1265,7 @@
 (define (tmtex-list l)
   (map-in-order tmtex l))
 
-(define (tmtex x)
+(tm-define (tmtex x)
   (if (string? x) (tmtex-string x)
       (tmtex-apply (car x) (cdr x))))
 
@@ -1080,27 +1274,29 @@
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 
 (drd-dispatcher tmtex-methods%
+  ((:or unknown uninit error raw-data) tmtex-noop)
   (document tmtex-document)
   (para tmtex-para)
   (surround tmtex-surround)
   (concat tmtex-concat)
-  (format tmtex-noop)
+  (group tmtex-group)
+  (hidden tmtex-noop)
   (hspace tmtex-hspace)
   (vspace* tmtex-noop)
   (vspace tmtex-vspace)
   (space tmtex-space)
   (htab tmtex-htab)
-  (split tmtex-noop)
   (move tmtex-noop)
   (resize tmtex-noop)
+  (repeat tmtex-noop)
   (float tmtex-float)
-  ((:or repeat datoms dlines dpages dbox) tmtex-noop)
+  ((:or datoms dlines dpages dbox) tmtex-noop)
+
   (with-limits tmtex-noop)
   (line-break tmtex-line-break)
   (new-line tmtex-new-line)
-  (line-sep tmtex-noop)
   (next-line tmtex-next-line)
-  (no_break tmtex-noop)
+  (no-break tmtex-no-break)
   (no-indent tmtex-no-first-indentation)
   (yes-indent tmtex-noop)
   (no-indent* tmtex-noop)
@@ -1113,7 +1309,7 @@
   (new-page tmtex-new-page)
   (new-dpage* tmtex-noop)
   (new-dpage tmtex-noop)
-  (group tmtex-group)
+
   (left tmtex-left)
   (mid tmtex-mid)
   (right tmtex-right)
@@ -1132,27 +1328,48 @@
   (neg tmtex-neg)
   (wide* tmtex-wide-star)
   ;;(tree tmtex-tree)
-  (tree tmtex-noop)
-  ((:or old-matrix old-table old-mosaic old-mosaic-item) tmtex-noop)
+  (tree tmtex-tree-eps)
+
   (tformat tmtex-tformat)
   ((:or twith cwith tmarker) tmtex-noop)
   (table tmtex-table)
   ((:or row cell subtable) tmtex-noop)
+
   (assign tmtex-assign)
   (with tmtex-with)
-  ((:or set reset) tmtex-noop)
-  (compound tmtex-compound)
-  ((:or begin end) tmtex-noop)
-  (include tmtex-noop)
-  ((:or macro func env eval) tmtex-noop)
+  (provides tmtex-noop)
   (value tmtex-compound)
-  (arg tmtex-noop)
-  ((:or backup quote delay hold release) tmtex-noop)
-  ((:or or xor and not plus minus times over div mod merge length range
-	number date translate is-tuple look-up equal unequal less lesseq
-	greater greatereq if case while extern authorize)
-   tmtex-noop)
-  ((:or inactive symbol latex hybrid tuple collection associate) tmtex-noop)
+  (quote-value tmtex-noop)
+  ((:or quote-value drd-props arg quote-arg) tmtex-noop)
+  (compound tmtex-compound)
+  ((:or xmacro get-label get-arity map-args eval-args mark eval) tmtex-noop)
+  ;; quote missing
+  (quasi tmtex-noop)
+  ;; quasiquote missing
+  ;; unquote missing
+  ((:or unquote* copy
+	if if* case while for-each
+	extern include use-package) tmtex-noop)
+
+  ((:or or xor and not plus minus times over div mod
+	merge length range number date translate change-case find-file
+	is-tuple look-up
+	equal unequal less lesseq greater greatereq) tmtex-noop)
+
+  ((:or cm-length mm-length in-length pt-length
+	bp-length dd-length pc-length cc-length
+	fs-length fbs-length em-length
+	ln-length sep-length yfrac-length ex-length
+	fn-length fns-length bls-length
+	spc-length xspc-length par-length pag-length
+	gm-length gh-length) tmtex-noop)
+
+  ((:or style-with style-with* style-only style-only*
+	active active* inactive inactive*
+	rewrite-inactive inline-tag open-tag middle-tag close-tag
+	symbol latex hybrid) tmtex-noop)
+
+  ((:or tuple attr tmlen collection associate backup) tmtex-noop)
   (label tmtex-label)
   (reference tmtex-reference)
   (pageref tmtex-pageref)
@@ -1160,34 +1377,70 @@
   (specific tmtex-specific)
   (hlink tmtex-hyperlink)
   (action tmtex-action)
-  ((:or tag meaning) tmtex-noop)
-  ((:or switch fold exclusive progressive superposed) tmtex-noop)
-  ((:or graphics point line arc bezier) tmtex-noop)
+  ((:or tag meaning flag) tmtex-noop)
+
+  ((:or anim-compose anim-repeat anim-constant
+	anim-translate anim-progressive video sound) tmtex-noop)
+
+  (graphics tmtex-graphics)
+  (superpose tmtex-noop)
+  ((:or gr-group gr-linear-transform
+	text-at cline arc carc spline spine* cspline fill) tmtex-noop)
   (postscript tmtex-postscript)
+  ((:or box-info frame-direct frame-inverse) tmtex-noop)
+
+  ((:or format line-sep split delay hold release
+	old-matrix old-table old-mosaic old-mosaic-item
+	set reset expand expand* hide-expand
+	apply begin end func env) tmtex-noop)
   
+  (shown tmtex-id)
   (!file tmtex-file)
   (!arg tmtex-tex-arg))
 
 (drd-table tmtex-tmstyle%
+  ((:or hide-preamble show-preamble) (,tmtex-default -1))
+  (hide-part (,tmtex-hide-part -1))
+  (show-part (,tmtex-show-part -1))
   (doc-data (,tmtex-doc-data -1))
+  ((:or doc-title doc-author-data doc-date doc-note
+	doc-keywords doc-AMS-class) (,tmtex-default -1))
+  ((:or author-name author-address author-note
+	author-email author-homepage) (,tmtex-default -1))
   (abstract (,tmtex-std-env 1))
   (appendix (,tmtex-appendix 1))
   ((:or theorem proposition lemma corollary proof axiom definition
-	notation conjecture remark note example exercise warning
-	convention quote quotation verse)
+	notation conjecture remark note example exercise problem warning
+	convention quote-env quotation verse)
    (,tmtex-std-env 1))
   ((:or verbatim code) (,tmtex-verbatim 1))
-  ((:or center indent body) (,tmtex-std-env 1))
+  ((:or center indent) (,tmtex-std-env 1))
   ((:or description itemize itemize-minus itemize-dot itemize-arrow
 	enumerate enumerate-numeric enumerate-roman enumerate-Roman
 	enumerate-alpha enumerate-Alpha)
    (,tmtex-list-env 1))
+  (really-tiny (,tmtex-tiny 1))
+  (very-tiny (,tmtex-tiny 1))
+  (really-small (,tmtex-scriptsize 1))
+  (very-small (,tmtex-scriptsize 1))
+  (smaller (,tmtex-footnotesize 1))
+  (flat-size (,tmtex-footnotesize 1))
+  (normal-size (,tmtex-normalsize 1))
+  (sharp-size (,tmtex-large 1))
+  (larger (,tmtex-Large 1))
+  (very-large (,tmtex-LARGE 1))
+  (really-large (,tmtex-LARGE 1))
+  (really-huge (,tmtex-Huge 1))
+
+  (math (,tmtex-math 1))
   ((:or equation equation*) (,tmtex-equation 1))
   ((:or eqnarray eqnarray* leqnarray*) (,tmtex-eqnarray 1))
+  (eq-number (,tmtex-default -1))
   ((:or the-index the-glossary) (,tmtex-dummy -1))
   ((:or table-of-contents) (,tmtex-toc 2))
   (bibliography (,tmtex-bib 4))
   (thebibliography (,tmtex-thebibliography 2))
+  (bib-list (,tmtex-second 2))
   (bibitem* (,tmtex-bibitem* -1))
   ((:or small-figure big-figure small-table big-table) (,tmtex-figure 2))
   (item (,tmtex-item 0))
@@ -1201,7 +1454,9 @@
   (choose (,tmtex-choose 2))
   ((:or strong em tt name samp abbr dfn kbd var acronym person)
    (,tmtex-modifier 1))
-  (menu (,tmtex-menu -1)))
+  (menu (,tmtex-menu -1))
+  (with-TeXmacs-text (,(tmtex-rename 'withTeXmacstext) 0))
+  (made-by-TeXmacs (,(tmtex-rename 'madebyTeXmacs) 0)))
 
 (drd-group tmtex-protected%
   a b c d i j k l o r t u v H L O P S
@@ -1209,20 +1464,89 @@
   ne ng ni nu oe or pi pm rm rq sb sc sf sl sp ss th to tt wd wp wr xi
   AA AE DH DJ Im NG OE Pi Pr Re SS TH Xi)
 
+;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
+;; Expansion of all macros which are not recognized by LaTeX
+;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
+
+(define tmtex-user-defs-table (make-ahash-table))
+
+(define (collect-user-defs-sub t)
+  (cond ((npair? t) (noop))
+	((macro-definition? t)
+	 (ahash-set! tmtex-user-defs-table (string->symbol (cadr t)) #t))
+	(else (for-each collect-user-defs-sub (cdr t)))))
+
+(define (collect-user-defs t)
+  (if (== (get-preference "texmacs->latex:expand-user-macros") "on") '()
+      (begin
+	(set! tmtex-user-defs-table (make-ahash-table))
+	(collect-user-defs-sub t)
+	(ahash-set->list tmtex-user-defs-table))))
+
+(define (as-string sym)
+  (with s (symbol->string sym)
+    (if (string-starts? s "begin-")
+	(substring s 6 (string-length s))
+	s)))
+
+(define (drd-first-list name)
+  (let* ((l1 (query (cons name '('first 'second))))
+	 (l2 (map (cut assoc-ref <> 'first) l1)))
+    (map as-string l2)))
+
+(define (tmtex-env-macro name)
+  `(associate ,name (xmacro "x" (eval-args "x"))))
+
+(tm-define (tmtex-env-patch t)
+  (let* ((l1 (drd-first-list 'tmtex-methods%))
+	 (l2 (drd-first-list 'tmtex-tmstyle%))
+	 (l3 (map as-string (drd-apply-list '(latex-tag%))))
+	 (l4 (map as-string (drd-apply-list '(latex-symbol%))))
+	 (l5 (list-difference l3 l4))
+	 (l6 (map as-string (collect-user-defs (tree->stree t))))
+	 (l7 (list-difference (list-union l2 (list-union l5 l6)) l1)))
+    `(collection ,@(map tmtex-env-macro l7))))
+
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;; Interface
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 
-(define (texmacs->latex x opts)
+(define (tmtex-get-style sty)
+  (cond ((string? sty) (set! sty (list sty)))
+	((func? sty 'tuple) (set! sty (cdr sty)))
+	((null? sty) (set! sty '("letter"))))
+  (if (== (car sty) "generic") (set! sty (cons "letter" (cdr sty))))
+  sty)
+
+(tm-define (texmacs->latex x opts)
   (if (tmfile? x)
       (let* ((body (tmfile-extract x 'body))
-	     (style* (tmfile-extract x 'style))
-	     (style (if (list? style*) style* (list style*)))
+	     (style (tmtex-get-style (tmfile-extract x 'style)))
+	     (main-style (or (tmtex-transform-style (car style)) "letter"))
 	     (lan (tmfile-init x "language"))
 	     (init (tmfile-extract x 'initial))
 	     (doc (list '!file body style lan init (get-texmacs-path))))
-	(texmacs->latex doc opts))
+	(latex-set-language lan)
+	(latex-set-style main-style)
+	(set! tmtex-chinese? (== lan "chinese"))
+	(set! tmtex-japanese? (== lan "japanese"))
+	(set! tmtex-korean? (== lan "korean"))
+	(set! tmtex-taiwanese? (== lan "taiwanese"))
+	(set! tmtex-oriental? (or tmtex-chinese? tmtex-japanese?
+				  tmtex-korean? tmtex-taiwanese?))
+	(with result (texmacs->latex doc opts)
+	  (set! tmtex-chinese? #f)
+	  (set! tmtex-japanese? #f)
+	  (set! tmtex-korean? #f)
+	  (set! tmtex-taiwanese? #f)
+	  (set! tmtex-oriental? #f)
+	  result))
       (let* ((x2 (tmtm-eqnumber->nonumber x))
 	     (x3 (tmtm-match-brackets x2)))
 	(tmtex-initialize opts)
-	(tmtex (tmpre-produce x3)))))
+	(with r (tmtex (tmpre-produce x3))
+	  (if (not tmtex-use-macros?)
+	      (set! r (latex-expand-macros r)))
+	  (if (not tmtex-use-catcodes?)
+	      (set! r (latex-expand-catcodes r)))
+	  r))))
